/*
 * $Id: //depot/prod/newport_news/main/modules/nn-codegen/src/main/java/com/netapp/common/logging/generator/MsgKeyGenerator.java#9 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.logging.generator;


import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.ClassUtils;

import com.netapp.common.generator.NetappClassGenerator;
import com.netapp.common.generator.NetappCodeFormatter;

/**
 * Construct the MsgKey enumerations for a given bundle file.
 * @author sunday
 * @version $Revision: #9 $ $Date: 2012/11/13 $
 */
public final class MsgKeyGenerator extends NetappClassGenerator {
    static final String BUNDLE_FILE_SUFFIX = ".properties";

    /** Class name of our MsgKey interface */
    private static final String MSG_KEY_INTERFACE = "com.netapp.common.logging.MessageKey";

    /** Class name of our MessageBundle class*/
    private static final String MESSAGE_BUNDLE_CLASS = "com.netapp.common.logging.MessageBundle";

    /** Class name of our Bundle annotation */
    private static final String BUNDLE_ANNOTATION = "com.netapp.common.logging.Bundle";

    /**
     * Regex for APPL-01234 codes.
     */
    private static final Pattern CODE_PATTERN = Pattern.compile("^(\\w+-\\d+):.*", Pattern.DOTALL);

    private static final NetappCodeFormatter FORMATTER;
    static {
        FORMATTER = new NetappCodeFormatter();
        FORMATTER.addHeaderLine("Generated by " + MsgKeyGenerator.class.getName());
    }

    protected String className;
    protected String packageName;
    protected String bundleName;
    protected ResourceBundle bundle;

    protected String classData;

    /**
     * Create a new generator for the supplied class and properties
     * bundle.
     *
     * @param qualifiedClassName
     * @param bundleName
     * @param classLoader
     * @exception LoggingException An error was detected in the supplied properties bundle.
     */
    public MsgKeyGenerator(String qualifiedClassName, String bundleName, ClassLoader classLoader) {
        super(ClassUtils.getPackageName(qualifiedClassName), FORMATTER);
        this.className = ClassUtils.getShortClassName(qualifiedClassName);

        if (MSG_KEY_INTERFACE.equals(qualifiedClassName)) {
            throw new RuntimeException("Generated class name may not be " + MSG_KEY_INTERFACE);
        }

        this.bundleName = (bundleName.endsWith(BUNDLE_FILE_SUFFIX)
                ? bundleName.substring(0, bundleName.length() - BUNDLE_FILE_SUFFIX.length())
                : bundleName)
            /*
             * On windows we need to convert the Windows path to unix path
             * so the generator will not use backslashes when referencing the
             * properties file location from the generated MsgKey enum.
             */
            .replaceAll("\\\\", "/");

        this.bundle = ResourceBundle.getBundle(this.bundleName, Locale.US, classLoader);
        this.generateMsgKeyClass();
    }

    public String getClassData() {
        return this.classData;
    }

    /**
     * Convert a java style camelCase string to an enumeration name
     * Convert to upper case, replacing '.' with '_' and adding a '_' before
     * uppercase letters.
     * @param input
     * @return String converted to Enum format
     */
    static String toEnumString(String input) {
        return input.substring(0, 1).toUpperCase()
            + input.substring(1).replaceAll("\\.|/|-", "_").replaceAll("([A-Z])", "_$1").toUpperCase();
    }

    /**
     * Generate the MsgKey enumeration into the current class buffer
     *
     * @exception LoggerException an error occured processing the input
     */
    private void generateMsgKeyClass() {
        addLine("@" + BUNDLE_ANNOTATION + "(\"" + bundleName + "\")");

        // Print our class name.
        // We always implement the interface by complete name,
        // as the user may name the resulting class MsgKey as well.
        addLine("public enum " + className + " implements " + MSG_KEY_INTERFACE + " {");
        increaseIndent();

        generateMsgEnumerations();

        // Generate the constructor and our interface implementation
        addLine();
        addLine("private final String propertyName;");
        addLine();
        addLine("private " + className + "(String propertyName) {");
        increaseIndent();
        addLine("this.propertyName = propertyName;");
        decreaseIndent();
        addLine("}");
        addLine();
        addLine("@Override");
        addLine("public String getPropertyName() {");
        increaseIndent();
        addLine("return this.propertyName;");
        decreaseIndent();
        addLine("}");

        addLine();
        addLine("/**");
        addLine(" * Localize this message key from its resource bundle.");
        addLine(" *");
        addLine(" * @param params Objects for variable substitution in string");
        addLine(" * @return Localized text string");
        addLine(" */");
        addLine("@Override");
        addLine("public String localize(Object... params) {");
        increaseIndent();
        addLine("return " + MESSAGE_BUNDLE_CLASS + ".localizeMessage(this, params);");
        decreaseIndent();
        addLine("}");

        addLine();
        addLine("/**");
        addLine(" * Find the " + className + " value for the given propertyName.");
        addLine(" *");
        addLine(" * @param propertyName the property name within the resource bundle for which to find the key");
        addLine(" * @return MsgKey value for the given propertyName");
        addLine(" * @throws IllegalArgumentException if no MsgKey value exists for the propertyName");
        addLine(" */");
        addLine("public static " + className + " parseKey(String propertyName) {");
        increaseIndent();
        addLine("for (" + className + " testKey : " + className + ".values()) {");
        increaseIndent();
        addLine("if (testKey.propertyName.equals(propertyName)) {");
        increaseIndent();
        addLine("return testKey;");
        decreaseIndent();
        addLine("}");
        decreaseIndent();
        addLine("}");
        addLine("throw new IllegalArgumentException(\"No enum found in \" + " + className + ".class.getName() + \" for property name \" + propertyName);");
        decreaseIndent();
        addLine("}");

        decreaseIndent();
        addLine("}");
    }

    /**
     *
     */
    private void generateMsgEnumerations() {
        /* Keep a list of message codes and keys we've seen already */
        Set<String> seenCodes = new HashSet<String>();
        Set<String> seenKeys = new HashSet<String>();

        // Iterate over the key message pairs
        // based on http://stackoverflow.com/questions/740299/how-do-i-sort-a-set-to-a-list-in-java
        List<String> bundleKeyList = new ArrayList<String>(bundle.keySet());
        Collections.sort(bundleKeyList);
        Iterator<String> bundleKeyIter = bundleKeyList.iterator();
        while (bundleKeyIter.hasNext()) {
            String bundleKey = bundleKeyIter.next();
            String comment = bundle.getString(bundleKey);
            String enumKey = toEnumString(bundleKey);

            if (seenKeys.contains(enumKey)) {
                throw new RuntimeException("Duplicate enumeration key " + enumKey);
            }
            seenKeys.add(enumKey);

            Matcher matcher = CODE_PATTERN.matcher(comment);
            if (matcher.matches()) {
                String code = matcher.group(1);
                if (seenCodes.contains(code)) {
                    throw new RuntimeException("Duplicate message code " + code);
                }
                seenCodes.add(code);
            }

            addJavadocLine(comment);

            addLine(enumKey + "(\"" + bundleKey + "\")" + (bundleKeyIter.hasNext() ? "," : ";"));
        }
    }

    /**
     * Convert a class name to a file path
     * @param clazz
     * @return Relative file path of the supplied class or package, excluding extension
     */
    public static String classNameToFilePath(String clazz) {
        return clazz.replace('.', File.separatorChar);
    }

    public static void main(String[] args) {
        if (args.length < 3) {
            printUsage();
        }

        try {
            String propFile = args[0];
            String outputDir = args[1];
            String className = args[2];
            generateMsgKey(propFile, new File(outputDir), className, MsgKeyGenerator.class.getClassLoader());
        } catch (RuntimeException e) {
            printUsage(e);
        } catch (IOException e) {
            printUsage(e);
        }
    }

    public static void generateMsgKey(String bundleName, File outputDir, String className, ClassLoader classLoader) throws IOException {
        File javaFile = calculateJavaFileName(outputDir, className);
        generateMsgKeyToFile(bundleName, javaFile, className, classLoader);
    }

    public static File calculateJavaFileName(File outputDir, String className) {
        String javaFileName = File.separator + classNameToFilePath(className) + ".java";
        return new File(outputDir, javaFileName);
    }

    @edu.umd.cs.findbugs.annotations.SuppressWarnings("DM_DEFAULT_ENCODING")
    public static void generateMsgKeyToFile(String bundleName, File javaFile, String className, ClassLoader classLoader) throws IOException {
        File javaFileDir = javaFile.getParentFile();
        if (!javaFileDir.exists() && !javaFileDir.mkdirs()) {
            throw new RuntimeException("Unable to create package directory " + javaFileDir);
        }

        MsgKeyGenerator generator = null;

        generator = new MsgKeyGenerator(className, bundleName, classLoader);

        Writer outputWriter = new FileWriter(javaFile);
        generator.writeCode(outputWriter);
        outputWriter.flush();
        outputWriter.close();
    }

    private static void printUsage() {
        System.err.println("Usage: MsgKeyGenerator <bundle_properties_file> <src_root_dir> <enum_class_name>");
    }

    private static void printUsage(Throwable t) {
        printUsage();
        if (t != null) {
            t.printStackTrace(System.err);
        }
        System.exit(1);
    }
}
